
#include "PDF.h"
#include <sys/stat.h>
#include "PDF_rep.h"

int PDF::file_length( const PDF &pdf ) {
  return pdf.rep->file_length1;
}

int PDF::offset_last_xref_table( const PDF &pdf ) {
  return pdf.rep->xref->getLastXRefPos();
}

PDF::Iref PDF::iref_catalog( const PDF &pdf ) {
  XRef *const xref = pdf.rep->xref;
  return Iref( xref->getRootNum(), xref->getRootGen() );
}

int PDF::n_obj( const PDF &pdf ) {
  return pdf.rep->n_obj1;
}

int PDF::offset( const PDF &pdf, const int i ) {
  return pdf.rep->xref->getEntry(i)->offset;
}

int PDF::n_page( const PDF &pdf ) {
  return pdf.rep->catalog2->getNumPages();
}

int PDF::i_page(
  const PDF &pdf,
  const Iref iref,
  const bool do_not_throw
) {
  const int i = pdf.rep->catalog2->findPage( iref.i, iref.gen );
  if (!do_not_throw && !i) throw PDF::Exc_PDF();
  return i;
}

PDF::Iref PDF::iref_page( const PDF &pdf, const int i ) {
  const Ref *const rp = pdf.rep->catalog2->getPageRef(i);
  if (!rp) throw PDF::Exc_PDF();
  return Iref( rp->num, rp->gen );
}

// The programmer does not feel sure that a little memory is not leaking
// here.  The amount of memory in question is small, and of course the
// system reclaims leaked memory at execution's end, anyway, so the leak
// if any is not serious; but even if not serious, leaking still is not
// neat.  The only documentation for Libpoppler appears to consist of
// its development headers, which seem insufficiently informative in the
// matter.  For these reasons, where in doubt, rather than risking
// improper deallocation, the code leaks.

PDF::PDF::PDF( const std::string &filename_pdf )
  : rep( new PDF_rep() )
{
  {
    struct stat s;
    if ( stat( filename_pdf.c_str(), &s ) ) throw Exc_IO();
    rep->file_length1 = s.st_size;
  }
  {
    GooString gs( filename_pdf.c_str() );
    rep->pdfdoc = new PDFDoc(&gs);
    if ( !rep->pdfdoc->isOk() ) throw Exc_IO();
  }
  {
    rep->xref = rep->pdfdoc->getXRef();
    if ( !rep->xref->isOk() ) throw Exc_PDF();
  }
  {
    Object *const obj = rep->xref->getTrailerDict();
    if ( !obj->isDict() ) throw Exc_PDF();
    rep->trailer = obj->getDict();
  }
  {
    Object obj;
    { char s[] = "Size"; rep->trailer->lookup( s, &obj ); }
    if ( !obj.isInt() ) throw Exc_PDF();
    rep->n_obj1 = obj.getInt();
  }
  {
    rep->catalog_obj = new Object();
    rep->xref->getCatalog( rep->catalog_obj );
    if ( !rep->catalog_obj->isDict() ) throw Exc_PDF();
    rep->catalog = rep->catalog_obj->getDict();
  }
  {
    rep->catalog2 = rep->pdfdoc->getCatalog();
    if ( !rep->catalog2->isOk() ) throw Exc_PDF();
  }
}

PDF::PDF::~PDF() {
  delete rep->catalog_obj;
  // For reasons this programmer does not understand, the Libpoppler PDFDoc
  // object does not seem to deallocate gracefully.  It is allowed to leak for
  // this reason.
  //delete rep->pdfdoc;
  delete rep;
}

PDF::PDF_rep *PDF::PDF::get_PDF_rep( const int magic ) {
  // The function demands a magic integer precisely to discourage
  // callers from calling it, and conversely to prevent it from
  // returning disruptive information to unwitting callers.  The integer
  // serves no other purpose.  Its value is not elsewhere documented.
  // If you must call this function, then supply the integer.  (The
  // integer's value has 1s in the zeroth and fifteenth bits, with six
  // more 1s scattered randomly across the fourteen places between.  It
  // has no significance.)
  return magic == 0x9f05 ? rep : 0;
}

